#include "stdafx.h"
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include "xs/xs_Core/xs_Float.h"
#include "xs/xs_Core/xs_Array.h"

//trying to make sure the optimizer doesn't get rid of it...
volatile long gk;



#define STRING_OUT(str)      printf (str);


typedef void* (*speedProc)(long numTests, float* rndbuf, long* intBuf, long iInc);


// ====================================================================================================================
//  FISTToInt
// ====================================================================================================================
__forceinline int FISTToInt (float f)
{
	int i;
	__asm fld [f]
	__asm fistp [i]
	return i;
}

// ====================================================================================================================
//  BitConvertToInt
// ====================================================================================================================
__forceinline int BitConvertToInt (float MyFloat)
{
   int MyInt;
   long FltInt = *(int *)&MyFloat;

   long mantissa = (FltInt & 0x07fffff) | 0x800000;
   long exponent = 150 - ((FltInt >> 23) & 0xff);

   if (exponent < 0)
     MyInt = (mantissa << -exponent);		      
   else
     MyInt = (mantissa >> exponent);

   if (FltInt & 0x80000000)
     MyInt = -MyInt;

   return MyInt;
}

// ====================================================================================================================
//  FloatTo23Bits
// ====================================================================================================================
int FloatTo23Bits(float x)
{
    float y = x + 1.f;
    return ((unsigned long&)y) & 0x7FFFFF;	// last 23 bits
}

// ====================================================================================================================
//  correctcheck
// ====================================================================================================================
int correctcheck(double scale,
                 char* label1, speedProc proc1, 
                 char* label2, speedProc proc2)
{
    char s[4096];
    enum { NUMTESTS = 10000 };
    
    //build a random# of doubles...
    static float rndbuf[NUMTESTS];
    static long ibuf1[NUMTESTS];
    static long ibuf2[NUMTESTS];
    int i;
    double minf,maxf;
    for (i=0; i<NUMTESTS; i++) 
        {
        double intg1 = rand()*scale;
        double intg2 = rand()*scale;
        double intg3 = rand()*scale;
        double frac = rand();

        rndbuf[i] = double(intg1) - double(intg2) + double(intg3) + double(frac) / RAND_MAX;
        if (i==0) 
            {
            minf = maxf = rndbuf[i];
            }
        else
            {
            if (minf>rndbuf[i]) minf=rndbuf[i];
            if (maxf<rndbuf[i]) maxf=rndbuf[i];
            }
        }

    //see if the conversion to int is the same as using our function....
    STRING_OUT("converting for buffer 1...\n");
    proc1(NUMTESTS, rndbuf, ibuf1, 1);
    STRING_OUT("converting for buffer 2...\n");
    proc2(NUMTESTS, rndbuf, ibuf2, 1);
    bool correct = true;
    STRING_OUT("checking buffer 1 against buffer 2\n");
    for (i=0; i<NUMTESTS; i++) 
        {
        if (ibuf1[i] != ibuf2[i]) 
            {
            correct = false;
            sprintf(s, "Bad: (sprintf)%.15g (%08x) -> (%s)%d != (%s)%d\n", 
                            rndbuf[i], *(int*)&rndbuf[i], 
                            label1, ibuf1[i], 
                            label2, ibuf2[i]);
            STRING_OUT(s);
            }
        }

    //report the range...
    sprintf(s, "%s vs %s %s, from %.15g to %.15g\n", 
                label1, label2,
                correct ? "--CORRECT--" : "*****INCORRECT*****", minf, maxf);
    STRING_OUT(s);
    return 0;
}


// ====================================================================================================================
//  Test loops
// ====================================================================================================================
#define LOOP_TEST(FUNC_NAME, PROC)      \
    void FUNC_NAME(long numTests, float* rndbuf, long* intBuf, long iInc)  \
            {for (long i=0; i<numTests; i++)      {float v=rndbuf[i]; *intBuf += PROC; intBuf+=iInc;}}

LOOP_TEST   (AnsiTest,         long(v));
LOOP_TEST   (SreeIntTest,      xs_ToInt(v));
LOOP_TEST   (SreeCRoundTest,   xs_CRoundToInt(v));
LOOP_TEST   (SreeRoundTest,    xs_RoundToInt(v));
LOOP_TEST   (FistTest,         FISTToInt(v));
LOOP_TEST   (BitTest,          BitConvertToInt(v));
LOOP_TEST   (BitTestSmall,     FloatTo23Bits(v));
LOOP_TEST   (FloorTest,        (long)floor(v));
LOOP_TEST   (SreeFloorTest,    (long)xs_FloorToInt(v));
LOOP_TEST   (AnsiFixedTest,    (long)(v*65536));
LOOP_TEST   (FistFixedTest,    (long)FISTToInt(v*65536));
LOOP_TEST   (SreeFixedTest,    (long)xs_Fix<16>::ToFix(v));

// ====================================================================================================================
//  runtest
// ====================================================================================================================
void runtest(char* label, speedProc proc, long trials, float* rndbuf, long numTests)
{
    char str[256];
    sprintf(str, "Testing %s... ", label);
    STRING_OUT(str);

    LARGE_INTEGER s1, s2, f;
	QueryPerformanceFrequency(&f);
	QueryPerformanceCounter(&s1);

    long myInt = 0;
	for (long i=0; i<trials; i++)
		proc(numTests, rndbuf, &myInt, 0);
    gk = myInt;


	QueryPerformanceCounter(&s2);
    real64 tmul = 1000./real64(f.QuadPart);
	real64 time = (s2.QuadPart - s1.QuadPart)*tmul;

    sprintf(str, "\t\tTime = %lg ms\n", time);
    STRING_OUT (str);
}



// ====================================================================================================================
//  speedcheck
// ====================================================================================================================
int speedcheck()
{
    enum { NUMTESTS = 10000000 };
    static float rndbuf[NUMTESTS];

    //random doubles
    int i;
    for (i=0; i<NUMTESTS; i++) 
        {
        int intg1 = rand();
        int intg2 = rand();
        int intg3 = rand();
        int frac = rand();
        rndbuf[i] = float(intg1) - float(intg2) + float(intg3) + float(frac) / RAND_MAX;
        }


    //time conversion
    runtest ("ANSI cast() ",     (speedProc)AnsiTest,        10, rndbuf, NUMTESTS);
    runtest ("SREE Int()  ",     (speedProc)SreeIntTest,     10, rndbuf, NUMTESTS);

    STRING_OUT ("\nRounding tests...\n");
    runtest ("BitConvert  ",     (speedProc)BitTest,         10, rndbuf, NUMTESTS);
    runtest ("fistp       ",     (speedProc)FistTest,        10, rndbuf, NUMTESTS);
    runtest ("SREE RoundInt()",  (speedProc)SreeRoundTest,   10, rndbuf, NUMTESTS);
    runtest ("BitConvert23",     (speedProc)BitTestSmall,    10, rndbuf, NUMTESTS);
    runtest ("SREE CRoundInt()", (speedProc)SreeCRoundTest,  10, rndbuf, NUMTESTS);

    STRING_OUT ("\nFixed tests...\n");
    runtest ("ANSI fixed()",    (speedProc)AnsiFixedTest,   10, rndbuf, NUMTESTS);
    runtest ("fistp fixed()",   (speedProc)FistFixedTest,   10, rndbuf, NUMTESTS);
    runtest ("Sree fixed()",    (speedProc)SreeFixedTest,   10, rndbuf, NUMTESTS);

    STRING_OUT ("\nFloor tests...\n");
    runtest ("ANSI floor()",    (speedProc)FloorTest,       10, rndbuf, NUMTESTS);
    runtest ("SREE Floor()",    (speedProc)SreeFloorTest,   10, rndbuf, NUMTESTS);

    return 0;
}




// ====================================================================================================================
//  main
// ====================================================================================================================
int _tmain(int argc, _TCHAR* argv[])
{
    real64 sval         = 1.25;
    int32 fix           = xs_Fix<16>::ToFix  (sval);
    int32 i             = xs_Fix<16>::ToInt  (fix);
    real64 r            = xs_Fix<16>::ToReal (fix);

    if (r!=sval)        {STRING_OUT ("bad real/fix...\n");}
    if (i!=floor(sval)) {STRING_OUT ("bad int/fix...\n");}

    STRING_OUT("starting float conversion testing...\n");
    speedcheck();

    STRING_OUT("\nstarting CORRECT testing...\n");
    correctcheck(10, "SREE CRound", (speedProc)SreeCRoundTest,
                     "fistp",       (speedProc)FistTest); 
    (void)getchar();
	return 0;
}

